package net.sf.seaf.test.internal;

import java.lang.annotation.Annotation;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Comparator;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashSet;

import javax.validation.constraints.Size;

import org.apache.commons.lang.reflect.FieldUtils;

public class InternalTestDataCreator {

	private final boolean recurse;

	private boolean rootLevel = true;

	public InternalTestDataCreator(boolean recurse) {
		this.recurse = recurse;
	}

	@SuppressWarnings("unchecked")
	public <T> T create(Class<T> type, Object baseValue, int index) {
		if (String.class.equals(type)) {
			return (T) (String.valueOf(baseValue) + prettyIntAsString(index));
		}
		if (byte[].class.equals(type)) {
			return (T) (String.valueOf(baseValue) + prettyIntAsString(index))
					.getBytes();
		}
		if (Boolean.class.equals(type)) {
			return (T) Boolean.valueOf(0 == (index % 2));
		}
		if (boolean.class.equals(type)) {
			return (T) (Object) Boolean.valueOf(0 == (index % 2))
					.booleanValue();
		}
		if (Long.class.equals(type)) {
			return (T) createLong(baseValue, index);
		}
		if (long.class.equals(type)) {
			return (T) (Object) createLong(baseValue, index).longValue();
		}
		if (Integer.class.equals(type)) {
			return (T) createInteger(baseValue, index);
		}
		if (int.class.equals(type)) {
			return (T) (Object) createInteger(baseValue, index).intValue();
		}
		if (Double.class.equals(type)) {
			return (T) new Double(
					createLong(baseValue, index).doubleValue() / 10);
		}
		if (double.class.equals(type)) {
			return (T) (Object) (createLong(baseValue, index).doubleValue() / 10);
		}
		if (Float.class.equals(type)) {
			return (T) new Float(
					createLong(baseValue, index).floatValue() / 100);
		}
		if (float.class.equals(type)) {
			return (T) (Object) (createLong(baseValue, index).floatValue() / 100);
		}
		if (Calendar.class.isAssignableFrom(type)) {
			return (T) createCalendar(baseValue, index);
		}
		if (Date.class.isAssignableFrom(type)) {
			return (T) createCalendar(baseValue, index).getTime();
		}
		if (Enum.class.isAssignableFrom(type)) {
			Object[] values = type.getEnumConstants();
			if (null == values) {
				return null;
			}
			int valueId = getOffset(baseValue, index) % values.length;
			return (T) values[valueId];
		}
		if (type.isArray()) {
			return (T) createArray(type.getComponentType(), baseValue, index);
		}
		if (!type.isInterface() && (rootLevel || recurse)) {
			rootLevel = false;
			return createBean(type, baseValue, index);
		}
		return null;
	}

	private Long createLong(Object baseValue, int index) {
		return new Long(Math.abs(repeatableHashCode(baseValue)) % 10000 * 1000
				+ prettyInt(index));
	}

	private Integer createInteger(Object baseValue, int index) {
		return createLong(baseValue, index).intValue() % 1000000;
	}

	/**
	 * Creates an array of 3 instances of the required type.
	 * 
	 * @param type
	 *            The type to instantiate and initialize
	 * @param baseValue
	 *            The base value for test data
	 * @param index
	 *            The offset for the test data
	 * @return An array of 3 initialized instances of the required type
	 */
	private <T> T[] createArray(Class<T> type, Object baseValue, int index) {
		@SuppressWarnings("unchecked")
		T[] a = (T[]) Array.newInstance(type, 3);
		a[0] = create(type, baseValue, index);
		rootLevel = true;
		a[1] = create(type, baseValue, index + 1);
		rootLevel = true;
		a[2] = create(type, baseValue, index + 2);
		return a;
	}

	/**
	 * Create an instance of the required non-simple type initialized with
	 * generated test data. The bean is instantiated by default constructor and
	 * all the available properties are set with test data.
	 * 
	 * @param type
	 *            The non-simple type to instantiate and initialize
	 * @param baseValue
	 *            The base value for test data
	 * @param index
	 *            The offset for the test data
	 * @return An initialized instance of the required type
	 */
	private <T> T createBean(Class<T> type, Object baseValue, int index) {
		try {
			T instance = type.newInstance();
			populateBean(instance, baseValue, index);
			return instance;
		} catch (Exception e) {
			throw new RuntimeException("Cannot instantiate class " + type, e);
		}
	}

	/**
	 * Set the available properties of a bean with test data.
	 * 
	 * @param bean
	 *            The bean whose properties should be set
	 * @param baseValue
	 *            The base value for test data
	 * @param index
	 *            The offset for the test data
	 */
	private void populateBean(Object bean, Object baseValue, int index) {
		index = getOffset(baseValue, index);
		Method[] allMethods = bean.getClass().getMethods();
		// must sort methods to ensure consistent results in different
		// environments:
		Arrays.sort(allMethods, new MethodComparator());
		for (int i = 0; i < allMethods.length; i++) {
			Method method = allMethods[i];
			String methodName = method.getName();
			Class<?>[] parameterTypes = method.getParameterTypes();
			String returnTypeName = method.getReturnType().toString();
			if (isSetter(methodName, parameterTypes, returnTypeName)) {
				Class<?> paramType = parameterTypes[0];
				String propertyName = extractPropertyName(methodName);
				Object value = create(paramType, propertyName, index);
				Field field = FieldUtils.getField(bean.getClass(),
						propertyName, true);
				if (null != value && null != field) {
					value = alignValueToValidations(value, field);
					setValue(bean, method, propertyName, value);
				}
			}
		}
	}

	/**
	 * Read JSR-303 validation annotations of the field and align the generated
	 * value according to them to ensure any validations of the bean pass.
	 * 
	 * @param value
	 *            Generated value
	 * @param field
	 *            Property field
	 * @return Value aligned with the validation annotations
	 */
	private Object alignValueToValidations(Object value, Field field) {
		if (value instanceof String && null != field) {
			Size size = extractAnnotation(field.getAnnotations(), Size.class,
					null);
			if (null != size) {
				String sValue = (String) value;
				int length = sValue.length();
				int max = size.max();
				if (length > max) {
					value = sValue.substring(length - max, length);
				}
			}
		}
		return value;
	}

	/**
	 * Extract annotation from an array of annotations. Dives into annotations
	 * of the annotations.
	 * 
	 * @param annotations
	 *            Array of annotations
	 * @param type
	 *            Annotation type to extract
	 * @return Extracted instance of the annotation or null if no such
	 *         annotation found
	 */
	private <T extends Annotation> T extractAnnotation(
			Annotation[] annotations, Class<T> type,
			HashSet<Annotation> processed) {
		if (null == processed)
			processed = new HashSet<Annotation>();
		for (int i = 0; i < annotations.length; i++) {
			Annotation annotation = annotations[i];
			if (!processed.contains(annotation)) {
				processed.add(annotation);
				if (type.isAssignableFrom(annotation.getClass())) {
					@SuppressWarnings("unchecked")
					T result = (T) annotation;
					return result;
				}
				T extracted = extractAnnotation(annotation.annotationType()
						.getAnnotations(), type, processed);
				if (null != extracted)
					return extracted;
			}
		}
		return null;
	}

	/**
	 * Simple method comparator.
	 */
	private class MethodComparator implements Comparator<Method> {
		public int compare(Method m1, Method m2) {
			return (m1.getDeclaringClass().getName() + m1.getName())
					.compareTo(m2.getDeclaringClass().getName() + m2.getName());
		}
	}

	/**
	 * Is a method a setter?
	 * 
	 * @param methodName
	 *            The name of the method
	 * @param parameterTypes
	 *            Types of parameters of the method
	 * @param returnTypeName
	 *            The return type of the method
	 * @return boolean Returns true iff the method is considered a setter
	 */
	private boolean isSetter(String methodName, Class<?>[] parameterTypes,
			String returnTypeName) {
		return methodName.startsWith("set") && 1 == parameterTypes.length;
		// Setters may return value, e.g. "this" to allow chaining.
		// && "void".equals(returnTypeName);
	}

	/**
	 * Extract the name of the property from the setter method name.
	 * 
	 * @param methodName
	 *            The setter method name
	 * @return String The name of the corresponding property
	 */
	private String extractPropertyName(String methodName) {
		return methodName.substring(3, 4).toLowerCase()
				+ methodName.substring(4);
	}

	/**
	 * Set the bean property value.
	 * 
	 * @param bean
	 *            The bean
	 * @param method
	 *            The setter method
	 * @param propertyName
	 *            The name of the property
	 * @param value
	 *            The value to set
	 */
	private void setValue(Object bean, Method method, String propertyName,
			Object value) {
		try {
			method.invoke(bean, new Object[] { value });
		} catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Cannot set bean property "
					+ propertyName);
		}
	}

	/**
	 * Create a test value of type Calendar.
	 * 
	 * @param baseValue
	 *            The base value for test data
	 * @param index
	 *            The offset for the test data
	 * @return An initialized instance of Calendar
	 */
	private Calendar createCalendar(Object baseValue, int index) {
		Calendar c = new GregorianCalendar(1950, 0, 1, 2, 3, 4);
		c.add(Calendar.MILLISECOND, prettyInt(getOffset(baseValue, index)));
		c.add(Calendar.SECOND, prettyInt(getOffset(baseValue, index)));
		c.add(Calendar.MINUTE, prettyInt(getOffset(baseValue, index)));
		c.add(Calendar.DAY_OF_MONTH, prettyInt(getOffset(baseValue, index)));
		c.add(Calendar.MONTH, prettyInt(getOffset(baseValue, index)));
		return c;
	}

	/**
	 * Calculate new offset from a base value and previous offset.
	 * 
	 * @param value
	 *            Base value
	 * @param i
	 *            Previous offset
	 * @return int New offset
	 */
	private int getOffset(Object value, int i) {
		return (Math.abs(repeatableHashCode(value)) + i) % 1000000;
	}

	/**
	 * Re-implement String hashCode() to ensure consistent results in future.
	 * 
	 * @param value
	 *            Value to be hashed
	 * @return Hash code of the value
	 */
	private int repeatableHashCode(Object value) {
		String s = String.valueOf(value);
		int h = 0;
		for (int i = 0; i < s.length(); i++) {
			h = 31 * h + s.charAt(i);
		}
		return h;
	}

	/**
	 * Cut an int to a pretty value.
	 * <p>
	 * Gets absolute value and last 3 digits.
	 * 
	 * @param i
	 *            An int
	 * @return int A pretty int value
	 */
	private int prettyInt(int i) {
		return Math.abs(i) % 1000;
	}

	/**
	 * Cut an int to a pretty String value.
	 * <p>
	 * Gets absolute value and last 3 digits. Pads left by zeroes if needed to
	 * always return 3 digits
	 * 
	 * @param i
	 * @return
	 */
	private String prettyIntAsString(int i) {
		int r = prettyInt(i);
		StringBuilder sb = new StringBuilder();
		if (r < 100)
			sb.append('0');
		if (r < 10)
			sb.append('0');
		sb.append(r);
		return sb.toString();
	}

}
