package net.sf.seaf.test.jmock;

import java.lang.reflect.ParameterizedType;

import org.jmock.Expectations;
import org.jmock.Mockery;
import org.jmock.Sequence;
import org.jmock.States;
import org.jmock.lib.AssertionErrorTranslator;
import org.jmock.lib.legacy.ClassImposteriser;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;

/**
 * JMockSupport provides support methods to use JMock more easily.
 * <p>
 * A test has basically the following steps:
 * <ol>
 * <li>Initialize JMock - <code>init()</code>.</li>
 * <li>Mock the dependencies - <code>mock()</code>.</li>
 * <li>Set expectations (i.e. what method calls on the mocks are expected) -
 * <code>check()</code>.</li>
 * <li>Call method under test.</li>
 * <li>Verify that the expectations were met - <code>verify()</code>.</li>
 * </ol>
 * 
 * <p>
 * When you extend JMockSupport the calls to <code>init()</code> and
 * <code>verify()</code> are performed automatically. The recommended usage
 * example is as follows:
 * 
 * <pre>
 * public class MyTest extends JMockSupport {
 * ...
 * 
 * ClassUnderTest classUnderTest = new ClassUnderTest();
 * Dependency dependency = mock(Dependency.class);
 * classUnderTest.setDependency(dependency);
 * 
 * check(new Expectations() {
 * 	{
 * 		one(dependency).calledMethod();
 * 	}
 * });
 * 
 * classUnderTest.methodUnderTest();
 * </pre>
 * <p>
 * When extending JMockSupport is not desirable, you can use JMockSupport as
 * follows:
 * 
 * <pre>
 * import static net.sf.seaf.test.util.JMockSupport.*;
 * 
 * init();
 * 
 * ClassUnderTest classUnderTest = new ClassUnderTest();
 * Dependency dependency = mock(Dependency.class);
 * classUnderTest.setDependency(dependency);
 * 
 * check(new Expectations() {
 * 	{
 * 		one(dependency).calledMethod();
 * 	}
 * });
 * 
 * classUnderTest.methodUnderTest();
 * 
 * verify();
 * </pre>
 */
public class JMockSupport {

	private static Mockery context;

	/**
	 * Initialize JMock {@link Mockery} context.
	 * <p>
	 * Extend your test class from JMockSupport to enable automatic
	 * initialization before each test method.
	 */
	public static void init() {
		context = new Mockery() {
			{
				// Enable mocking of classes (not just interfaces):
				setImposteriser(ClassImposteriser.INSTANCE);
				setExpectationErrorTranslator(AssertionErrorTranslator.INSTANCE);
			}
		};
	}

	/**
	 * Return JMock context, if there is any.
	 * <p>
	 * Otherwise throw {@link IllegalStateException}.
	 * 
	 * @return JMock context
	 * @throws IllegalStateException
	 *             when the JMock context is not initialized
	 */
	private static final Mockery getContext() {
		if (null == context) {
			throw new IllegalStateException(
					"JMock is not initialized. Did you call init()?");
		}
		return context;
	}

	/**
	 * Create a mock instance of the required type.
	 * <p>
	 * Use when there is only one mock instance of the given class in the test,
	 * in case there are multiple instances of the same class use
	 * <code>mock(Class<T> typeToMock, String name)</code> instead.
	 * 
	 * @param typeToMock
	 *            Class to be mocked
	 * @return An mock instance of the requested class
	 */
	public static final <T> T mock(Class<T> typeToMock) {
		return getContext().mock(typeToMock);
	}

	/**
	 * Create a mock instance of the required type with a given unique name.
	 * <p>
	 * Names the mock instance uniquely within the test to distinguish multiple
	 * mock instances of the same class.
	 * 
	 * @param typeToMock
	 *            Class to be mocked
	 * @param name
	 *            A name given to the returned instance
	 * @return An mock instance of the requested class
	 */
	public static final <T> T mock(Class<T> typeToMock, String name) {
		return getContext().mock(typeToMock, name);
	}

	/**
	 * Creates a JMock {@link Sequence} to be used to order expectations.
	 * 
	 * @param name
	 *            Name of the sequence
	 * @return Sequence
	 */
	public static final Sequence createSequence(String name) {
		return getContext().sequence(name);
	}

	/**
	 * Creates a JMock {@link States} to be used to relate expectations to a
	 * state machine.
	 * 
	 * @param name
	 *            Name of the state machine
	 * @return States
	 */
	public static final States createStates(String name) {
		return getContext().states(name);
	}

	/**
	 * Declare test expectations to be checked.
	 * <p>
	 * Reset JMock context for next test execution.
	 * 
	 * @param expectations
	 *            Test expectations
	 */
	public static final void check(Expectations expectations) {
		getContext().checking(expectations);
	}

	/**
	 * Unwrap {@link Error} possibly thrown from a void method invocation.
	 * 
	 * @param voidInvoker
	 *            Void method invoker, wraps the call to the void method
	 */
	public static final void unWrapError(VoidInvoker voidInvoker) {
		try {
			voidInvoker.invoke();
		} catch (Throwable t) {
			unWrapAndThrowError(t);
		}
	}

	/**
	 * Unwrap {@link Error} possibly thrown from a method invocation that
	 * returns a value.
	 * 
	 * @param returningInvoker
	 *            Returning method invoker, wraps the call to the method
	 *            returning a value
	 */
	public static final <T> T unWrapError(ReturningInvoker<T> returningInvoker) {
		try {
			return returningInvoker.invoke();
		} catch (Throwable t) {
			unWrapAndThrowError(t);
			// redundant:
			throw (RuntimeException) t;
		}
	}

	/**
	 * Invocation wrapper around a void method.
	 */
	public interface VoidInvoker {
		/**
		 * Wrap invocation of a void method.
		 */
		void invoke();
	}

	/**
	 * Invocation wrapper around a method returning a value.
	 */
	public interface ReturningInvoker<T> {
		/**
		 * Wrap invocation of a method returning a value.
		 */
		T invoke();
	}

	/**
	 * Unwrap {@link Error} if there is any and throw it, otherwise throw as
	 * {@link RuntimeException}.
	 * 
	 * @param t
	 */
	private static final void unWrapAndThrowError(Throwable t) {
		Throwable cause = t;
		while (null != cause.getCause()) {
			cause = cause.getCause();
		}
		if (cause instanceof Error) {
			throw (Error) cause;
		}
		throw (RuntimeException) t;
	}

	/**
	 * Invocation wrapper around exception verification.
	 * 
	 * @param <E>
	 *            Type of exception to verify
	 */
	public interface ExceptionVerifier<E extends Exception> extends VoidInvoker {
		/**
		 * Verify the exception.
		 * 
		 * @param e
		 *            Exception caught from business method invocation
		 */
		void verify(E e);
	}

	/**
	 * Invoke business method, un-wrap an exception thrown from it and verify
	 * the un-wrapped exception.
	 * 
	 * @param <E>
	 *            The (un-wrapped) exception type expected
	 * @param verifier
	 *            Exception verifier
	 */
	public static final <E extends Exception> void verifyUnWrappedException(
			ExceptionVerifier<E> verifier) {
		try {
			unWrapError(verifier);
			Assert.fail("Should fail");
		} catch (Exception e) {
			if (getExceptionClass(verifier).isAssignableFrom(e.getClass())) {
				@SuppressWarnings("unchecked")
				E exception = (E) e;
				verifier.verify(exception);
			} else {
				// bypass Assert.fail to init cause:
				throw new AssertionError(e);
			}
		}
	}

	/**
	 * Extract the exception type from the verifier.
	 * 
	 * @param verifier
	 *            Verifier
	 * @return Exception type of the verifier
	 */
	private static final Class<?> getExceptionClass(
			ExceptionVerifier<?> verifier) {
		return (Class<?>) ((ParameterizedType) verifier.getClass()
				.getGenericInterfaces()[0]).getActualTypeArguments()[0];
	}

	/**
	 * Verify the expectations were met.
	 * <p>
	 * Extend your test class from JMockSupport to enable automatic verification
	 * after each test method.
	 */
	public static final void verify() {
		getContext().assertIsSatisfied();
		context = null;
	}

	/**
	 * Extend your test class from JMockSupport to enable automatic
	 * initialization of JMock before each test method in the test class.
	 */
	@Before
	public final void initBeforeEachTest() {
		init();
	}

	/**
	 * Extend your test class from JMockSupport to enable automatic verification
	 * of expectations after each test method in the test class.
	 */
	@After
	public final void verifyAfterEachTest() {
		verify();
	}

}
