package net.sf.seaf.test.util;

import static org.junit.Assert.*;

import org.jmock.Expectations;
import org.jmock.Mockery;
import org.jmock.Sequence;
import org.jmock.lib.AssertionErrorTranslator;
import org.jmock.lib.legacy.ClassImposteriser;

/**
 * Template class for JMock-based JUnit tests.
 * <p>
 * Subclass it and implement behavior specific to the business method under
 * test; in particular, the following methods are required to be implemented:
 * <ul>
 * <li>setUp() - set the test up, typically mock-up test objects,
 * <li>getExpectations() - provide JMock Expectations (what the method under
 * test is expected to perform),
 * <li>executeBusinessMethod() - the call to the method under test itself.
 * </ul>
 * In addition, the following methods can also be overridden as required by the
 * test:
 * <ul>
 * <li>beforeSetUp() - test-case specific set up (as opposed to setUp(), where
 * the common setup shared by all the test-cases of a particular business method
 * are placed) performed before the common set up,
 * <li>afterSetUp() - test-case specific set up performed after the common set
 * up,
 * <li>getSpecificExpectations() - test-case specific JMock Expectations,
 * <li>afterBusinessMethod() - processing done immediately after the business
 * method has been executed,
 * <li>exPost() - processing done at the very end of the test, typically
 * additional JUnit assertions,
 * <li>verifyException() - override to validate exception thrown from the
 * business method; by default fails the test.
 * </ul>
 * The test is triggered by calling the doTest() method. This method implements
 * the standard algorithm of a JMock test, delegating to template methods listed
 * above.
 */
public abstract class TestTemplate {

	/**
	 * JMock mockery context.
	 */
	public final Mockery context;

	/**
	 * Default constructor.
	 * <p>
	 * Instantiates the context. Uses a class imposterizer - even classes can be
	 * mocked, not just interfaces.
	 */
	public TestTemplate() {
		context = new Mockery() {
			{
				// Enable mocking of classes (not just interfaces):
				setImposteriser(ClassImposteriser.INSTANCE);
				setExpectationErrorTranslator(AssertionErrorTranslator.INSTANCE);
			}
		};
	}

	/**
	 * Shortcut for context.mock() method.
	 * 
	 * @param typeToMock
	 *            Class to be mocked
	 * @return An mock instance of the requested class
	 */
	public final <T> T mock(Class<T> typeToMock) {
		return context.mock(typeToMock);
	}

	/**
	 * Shortcut for context.mock() method.
	 * <p>
	 * Names the returned instance. To be used to distinguish multiple mock
	 * instances of the same class.
	 * 
	 * @param typeToMock
	 *            Class to be mocked
	 * @param name
	 *            A name given to the returned instance
	 * @return An mock instance of the requested class
	 */
	public final <T> T mock(Class<T> typeToMock, String name) {
		return context.mock(typeToMock, name);
	}

	/**
	 * Shortcut for context.sequence() method. Creates a JMock Sequence to be
	 * used to order expectations.
	 * 
	 * @param name
	 *            Name of the sequence
	 * @return Sequence
	 */
	public final Sequence createSequence(String name) {
		return context.sequence(name);
	}

	/**
	 * Override to provide test case-specific set up before the common,
	 * method-specific set up.
	 */
	public void beforeSetUp() {
	}

	/**
	 * Override to provide method-specific set up.
	 */
	public abstract void setUp();

	/**
	 * Override to provide test case-specific set up after the common,
	 * method-specific set up.
	 */
	public void afterSetUp() {
	}

	/**
	 * Override to provide method-specific JMock expectations.
	 * 
	 * @return Expectations
	 * @throws Exception
	 *             Should never be thrown.
	 */
	public abstract Expectations getExpectations() throws Exception;

	/**
	 * Override to provide test case-specific JMock expectations.
	 * 
	 * @return Expectations
	 * @throws Exception
	 *             Should never be thrown.
	 */
	public Expectations getSpecificExpectations() throws Exception {
		return new Expectations();
	}

	private void createExpectations() throws Exception {
		context.checking(getExpectations());
		context.checking(getSpecificExpectations());
	}

	/**
	 * Override to call the business method tested. No other test logic should
	 * be included.
	 * 
	 * @throws Exception
	 *             Thrown as requested by the test case.
	 */
	public abstract void executeBusinessMethod() throws Exception;

	/**
	 * Override to perform specific behavior immediately after the business
	 * method has been called.
	 * 
	 * @throws Exception
	 *             Thrown as requested. Will be processed by verifyException().
	 */
	public void afterBusinessMethod() throws Exception {
	}

	private void assertIsSatisfied() {
		context.assertIsSatisfied();
	}

	/**
	 * Handles an exception thrown from the business method.
	 * <p>
	 * Default implementation fails the test. Override to handle the exception
	 * as required by the test case.
	 * 
	 * @param e
	 *            Exception The exception thrown from the business method.
	 */
	public void verifyException(Exception e) {
		e.printStackTrace();
		fail("Business method should not throw exceptions");
	}

	/**
	 * Override to perform specific behavior at the end of the test.
	 * <p>
	 * Typically, additional assertions come here.
	 */
	public void exPost() {
	}

	/**
	 * Performs the test. To be called from the main JUnit test code.
	 */
	public final void doTest() {
		beforeSetUp();
		setUp();
		afterSetUp();
		try {
			createExpectations();
		} catch (Exception e) {
			e.printStackTrace();
			fail("Expectations must not throw exceptions");
		}
		try {
			executeBusinessMethod();
			afterBusinessMethod();
		} catch (Exception e) {
			verifyException(e);
		}
		assertIsSatisfied();
		exPost();
	}

}
